/***
*** hardware.c
***
***  General description of this file:
***     Device driver source code for General Standards 16AO12
***     family of analog I/O boards. This file is part of the Linux
***     driver source distribution for this board.
***
***  Copyrights (c):
***     General Standards Corporation (GSC), 2005-2006
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC. 
***
***  Platform (tested on, may work with others):
***     Linux, kernel version 2.4.18+, 2.6.x,  Red Hat distribution, Intel hardware.
***/

//////////////////////////////////////////////////////////////////////////
// set the following flag to trace interrupt debug messages.
#ifdef DEBUG
#define TRACE_LOCAL TRUE
#endif

#include "internals.h"

/************************************************************************/
/*                                                                      */
/************************************************************************/
void DisableIrqPlx(struct device_board *device)
{
    andPLX(device, INT_CTRL_STATUS, ~(IRQ_PCI_ENABLE | IRQ_DMA_0_ENABLE | IRQ_DMA_1_ENABLE | IRQ_LOCAL_PCI_ENABLE));
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void EnableIrqPlxLocal(struct device_board *device)
{
    orPLX(device, INT_CTRL_STATUS, IRQ_PCI_ENABLE | IRQ_LOCAL_PCI_ENABLE);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void EnableIrqPlxDMA0(struct device_board *device)
{
    orPLX(device, INT_CTRL_STATUS, IRQ_PCI_ENABLE | IRQ_DMA_0_ENABLE);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void EnableIrqPlxDMA1(struct device_board *device)
{
    orPLX(device, INT_CTRL_STATUS, IRQ_PCI_ENABLE | IRQ_DMA_1_ENABLE);
}

/************************************************************************/
/* writeLocal                                                           */
/*                                                                      */
/* Write to the board local registers, whether the local registers      */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
void writeLocal(struct device_board *device, unsigned address, __u32 value)
{
    unsigned long flags;

    spin_lock_irqsave(&device->regLock,flags);
    if (device->PciBar[2].IsIoMapped) /* i/o space */
    {
        //msg("writeLocal I/O space addr: %.8X value: %.8X.\n",address,value);
        outl(value, (unsigned) device->local_addr+address*4);
    }
    else
    {
        writel(value, device->local_addr+address);
    }
    spin_unlock_irqrestore(&device->regLock,flags);
};

/************************************************************************/
/* andLocal                                                           */
/*                                                                      */
/* Write to the board local registers, whether the local registers      */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
void andLocal(struct device_board *device, unsigned address, __u32 value)
{
    unsigned long flags;
    unsigned long regval;

    spin_lock_irqsave(&device->regLock,flags);
    if (device->PciBar[2].IsIoMapped) /* i/o space */
    {
        regval = inl((unsigned) device->local_addr+address*4);
        regval &=value;
        //msg("writeLocal I/O space addr: %.8X value: %.8X.\n",address,value);
        outl(regval, (unsigned) device->local_addr+address*4);
    }
    else
    {
        regval=readl(device->local_addr+address);
        regval &= value;
        writel(regval, device->local_addr+address);
    }
    spin_unlock_irqrestore(&device->regLock,flags);
};

/************************************************************************/
void orLocal(struct device_board *device, unsigned address, __u32 value)
{
    unsigned long flags;
    unsigned long regval;

    spin_lock_irqsave(&device->regLock,flags);
    if (device->PciBar[2].IsIoMapped) /* i/o space */
    {
        regval = inl((unsigned) device->local_addr+address*4);
        regval |=value;
        //msg("writeLocal I/O space addr: %.8X value: %.8X.\n",address,value);
        outl(regval, (unsigned) device->local_addr+address*4);
    }
    else
    {
        regval=readl(device->local_addr+address);
        regval |= value;
        writel(regval, device->local_addr+address);
    }
    spin_unlock_irqrestore(&device->regLock,flags);
};

/************************************************************************/
/* writePLX                                                             */
/*                                                                      */
/* Write to the board local registers, whether the local registers      */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
void writePLX(struct device_board *device, unsigned address, unsigned long value)
{
    //    unsigned long flags;

    //msg("writePLX base addr: %p offset: %x reg: %x value: %.8lx\n", device->runtime_addr,address,address*4,value);

    //    spin_lock_irqsave(&device->regLock,flags);
    writel(value, device->runtime_addr+address);
    //    spin_unlock_irqrestore(&device->regLock,flags);
};

/************************************************************************/
/* andPLX                                                             */
/*                                                                      */
/* Write to the board local registers, whether the local registers      */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
void andPLX(struct device_board *device, unsigned address, unsigned long value)
{
    unsigned long flags;
    unsigned long regval;

    //msg("andPLX base addr: %p offset: %x reg: %x value: %.8lX\n", device->runtime_addr,address,address*4,value);
    spin_lock_irqsave(&device->regLock,flags);
    regval = readl(device->runtime_addr+address);
    regval &= value;
    writel(regval, device->runtime_addr+address);
    spin_unlock_irqrestore(&device->regLock,flags);
};

/************************************************************************/
/* orPLX                                                                */
/*                                                                      */
/* Write to the board local registers, whether the local registers      */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
void orPLX(struct device_board *device, unsigned address, unsigned long value)
{
    unsigned long flags;
    unsigned long regval;

    //msg("orPLX base addr: %p offset: %x reg: %x value: %.8lX\n", device->runtime_addr,address,address*4,value);
    spin_lock_irqsave(&device->regLock,flags);
    regval = readl(device->runtime_addr+address);
    regval |= value;
    writel(regval, device->runtime_addr+address);
    spin_unlock_irqrestore(&device->regLock,flags);
};

/************************************************************************/
/* readPLX                                                             */
/*                                                                      */
/* Write to the board local registers, whether the local registers      */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
__u32 readPLX(struct device_board *device, unsigned address)
{
    return readl(device->runtime_addr+address);
};

/************************************************************************/
/* readLocal                                                            */
/*                                                                      */
/* Read from the board local registers, whether the local registers     */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
unsigned readLocal(struct device_board *device, unsigned address)
{
    if (device->PciBar[2].IsIoMapped) /* i/o space */
    {
        //msg("readLocal I/O space addr: %.8X \n",address);
        return inl((unsigned)device->local_addr+address*4);
    }
    else
    {
        return readl(device->local_addr+address);
    }
    return 0;
};

/************************************************************************/
/*                                                                      */
/************************************************************************/
void DisableIrqLocalAll(struct device_board *device)
{

    andLocal(device, BOARD_CTRL_REG, ~(BCR_IRQ_MASK));
    andLocal(device, BOARD_CTRL_REG, ~(BCR_INTERRUPT_REQUEST_FLAG));
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void EnableIrqLocal(struct device_board *device, u32 irq)
{
    u32 regval;

    regval = readLocal(device,BOARD_CTRL_REG);
    regval &= (~(BCR_IRQ_MASK));
    regval |= (irq << BCR_IRQ_SHIFT);
    writeLocal(device,regval,BOARD_CTRL_REG);
}

/************************************************************************/
/*                                                                      */
/************************************************************************/
void DisableIrqLocal(struct device_board *device, u32 irq)
{
    u32 regval;

    regval = readLocal(device,BOARD_CTRL_REG);
    regval &= (~(BCR_IRQ_MASK));
    writeLocal(device,regval,BOARD_CTRL_REG);
    regval &= (~(BCR_INTERRUPT_REQUEST_FLAG));
    writeLocal(device,regval,BOARD_CTRL_REG);
}

