/***
*** internals.h
***
***  General description of this file:
***     Device driver source code for General Standards 16AO family of
***     16-bit analog output boards. This file is part of the Linux
***     driver source distribution for this board.
***
***  Copyrights (c):
***     General Standards Corporation (GSC), 2003-2006
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC.
***
***  Platform (tested on, may work with others):
***     Linux, kernel version 2.4.x, Red Hat distribution, Intel hardware.
***/

#ifndef PCI16AO12_H
#define PCI16AO12_H

// include the ioctl file here, since it is used for all files->

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/pci.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/ioctl.h>
#include <linux/spinlock.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "sysdep.h"

#include "gsc16ao12_ioctl.h"

#define DRIVER_VERSION "1.9.3"
#define COPYRIGHT " 2004-2006 "

/* module name NOTE:  This name must be the same as the driver binary.  The install script
uses the name of the binary name determine if the driver is installed properly, and the
driver uses this name to announce itself.  The name of the binary is set in Makefile.
*/
#define GSC_NAME "gsc_16ao12"

#define GSC_SUBVENDOR PCI_VENDOR_ID_PLX

#define PCI_NUM_BARS 6

#define MAX_BOARDS 10 /* maximum number of boards that driver supports. Arbitrarily chosen,
change if you can actully support more. */


struct board_entry {
    int subsystem_vendor;
    int subsystem_device;
    char name[40];
    int index;
};

// macro used to write to the mailbox register to trace execution with
// a bus analyzer.

#undef mark            /* undef it, just in case */
#ifdef DEBUG
#define mark(x) writePLX(device, MAILBOX_REGISTER_0, x)
#else
#define mark(x)
#endif

#undef msg             /* undef it, just in case */
#ifdef TRACE_LOCAL
#  ifdef __KERNEL__
/* This one if debugging is on, and kernel space */
#    define msg(fmt, args...) printk( KERN_DEBUG GSC_NAME " " fmt, ## args)
#  else
/* This one for user space */
#    define msg(fmt, args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#  define msg(fmt, args...) /* not debugging: nothing */
#endif
#    define errmsg(fmt, args...) printk( KERN_ERR GSC_NAME " " fmt, ## args)
#    define logmsg(fmt, args...) printk( KERN_INFO GSC_NAME " " fmt, ## args)

#if !defined(U64)
typedef union _U64
{
    struct
    {
        u32  LowPart;
        u32  HighPart;
    }u;

    u64 QuadPart;
} U64;
#endif

// PCI BAR Space information 
typedef struct _PCI_BAR_INFO
{
    u32     *pVa;                        // BAR Kernel Virtual Address
    U64      Physical;                   // BAR Physical Address
    u32      Size;                       // BAR size
    int  IsIoMapped;                 // Memory or I/O mapped?
} PCI_BAR_INFO;

enum {
    TX
};

#define EVENT_INIT_COMPLETE             0
#define EVENT_AUTOCAL_COMPLETE          1
#define EVENT_OUT_BUFFER_EMPTY          2
#define EVENT_OUT_BUFFER_LOW_QUARTER    3  
#define EVENT_OUT_BUFFER_HIGH_QUARTER   4 
#define EVENT_BURST_TRIGGER_READY       5
#define EVENT_LOAD_READY                6
#define EVENT_END_LOAD_READY            7      
#define EVENT_DMA_PENDING               8
#define EVENT_LAST_IRQ_EVENT EVENT_DMA_PENDING+1

struct dma_data_struct {
    u32 dmaState;
    u32 *intermediateBuffer;        // DMA buffer 
    u32 intermediatePhysicalAddr;   // physical address of above 
    u32 userOffset;                 // current offset into user buffer.
    u32 userRemaining;              // space remaining in the user buffer.
    char *userBuffer;               // pointer to the current user buffer.
    u32 intermediateSamples;        // unread samples in the DMA buffer 
    u32 intermediateStart;          // first unread sample in the DMA buffer 
    u32 dmasamples;                 // unread samples in the DMA buffer 
    u32 dmastart;                   // first unread sample in the DMA buffer 
};

/* card descriptor structure */
struct device_board {
    struct pci_dev *pdev;           /* the kernel PCI device descriptor */
    u32 *runtime_addr;    /* remapped runtime register address space */
    u32 *local_addr;      /* remapped local register address space */
    long timeout;                   /* timeout flag */
    int timeout_seconds;            /* time to wait before timeout (in seconds) */
    int busy;                       /* device opened */
    int irqlevel;                   /* interrupt level */
    int minor;                      /* device minor number */
    int error;                      /* error code */
    int signalno;                   /* signal index if user requested notification */
    int signalev;                   /* signal the user if this event occurs */
    volatile atomic_t irq_event_pending[EVENT_LAST_IRQ_EVENT+1]; // array of flags for what irq is active

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,0)
    struct wait_queue *ioctlwq;     /* queue of tasks sleeping in ioctl */
    struct wait_queue *writewq;      /* queue of tasks sleeping in write */
    struct wait_queue *dmawq;       /* queue of tasks sleeping on DMA transfer completion */
    struct wait_queue *event_queue_waiting[EVENT_LAST_IRQ_EVENT+1];
#else
    wait_queue_head_t ioctlwq;      /* queue of tasks sleeping in ioctl */
    wait_queue_head_t writewq;          /* queue of tasks sleeping in write */
    wait_queue_head_t dmawq;        /* queue of tasks sleeping on DMA transfer completion */
    wait_queue_head_t *event_queue_waiting[EVENT_LAST_IRQ_EVENT+1];
#endif

    int board_type; /* an index for which board type for drivers that support multiple types. */
    int board_index;
    int dmaMode;
    struct dma_data_struct dma_data[2];
    spinlock_t regLock;

#ifdef DEBUG
    int nread;
    int last;
#endif

    int debug_state;
    struct semaphore sem;
    struct file *fp;                // file pointer for user data.

    int fillBuffer;
    struct timer_list watchdog_timer;
    PCI_BAR_INFO           PciBar[PCI_NUM_BARS];    // PCI BARs information
    struct device_board *next;             /* next board in linked list */
};

#define waitEvent(event)    wait_event_interruptible(*device->event_queue_waiting[event],(!atomic_read(&device->irq_event_pending[event])))
#define setEvent(event)     atomic_set(&device->irq_event_pending[event],TRUE)
#define clearEvent(event)   atomic_set(&device->irq_event_pending[event],FALSE)


/* non-static function prototypes */
int device_ioctl(struct inode *inode, struct file *fp, unsigned int num, unsigned long arg);

/* timing and delay stuff */
#define INIT_TIMEOUT_MS         2000        /* initialization timeout */
#define AUTOCAL_TIMEOUT_MS      5000        /* auto calibration timeout */
#define CH_READY_TIMEOUT_MS     500         /* timeout after channel setup operations */
#define MSECS_TO_SLEEP(MS)      ((MS * HZ) / 1000)

/* additional, internally used interrupt event codes */
#define NO_EVENT                (-1)

/* DMA buffer size (16 kBytes on Intel) */
#define DMA_ORDER               4
#define DMA_SIZE                (PAGE_SIZE << DMA_ORDER)
#define DMA_SAMPLES             (DMA_SIZE / sizeof(long))

/* DMA command codes */

#define START_DMA_CMD_0         0x0B
#define START_DMA_CMD_1         0xB00
#define STOP_DMA_CMD_0_MASK     0xFF01
#define STOP_DMA_CMD_1_MASK     0x00FF
#define NON_DEMAND_DMA_MODE     0x00020D43
#define DMA_CMD_STAT_INT_CLEAR  0x08
#define PCI_INT_ENABLE          0x00050900

// global data.  Only created during init.
extern s8 built[128];
extern int num_boards;
extern int proc_enabled;
extern struct board_entry boards_supported[];
extern const char *event_name[];

#ifdef DEBUG
extern u32 int_other_count[MAX_BOARDS];
extern u32 int_count[MAX_BOARDS];
extern u32 dma_count[MAX_BOARDS];
extern u32 channel_irq[MAX_BOARDS];
extern u32 channel_expected[MAX_BOARDS];
extern int board_type[MAX_BOARDS];
extern void * context[MAX_BOARDS];
#endif

//functions
void writeLocal(struct device_board *, unsigned address, unsigned value);
void orLocal(struct device_board *,unsigned value, unsigned address);
void andLocal(struct device_board *,unsigned value, unsigned address);
unsigned readLocal(struct device_board *,unsigned address);
int read_proc(char* page, char** start, off_t offset, int count, int* eof, void* data);
int proc_get_info(char* page, char** start, off_t offset, int count, int dummy);
int init_module(void);
void cleanup_module(void);
int device_open(struct inode *inode, struct file *fp);
int device_close(struct inode *inode, struct file *fp);
int device_write(struct file *fp, const char *buf, size_t size, loff_t * lt);
int device_read(struct file *fp, char *buf, size_t size, loff_t * lt);
void timeout_handler(unsigned long ptr);

void DisableIrqPlx(struct device_board *device);
void EnableIrqPlxLocal(struct device_board *device);
void EnableIrqPlxDMA0(struct device_board *device);
void EnableIrqPlxDMA1(struct device_board *device);
void DisableIrqLocalAll(struct device_board *device);

void EnableIrqLocal(struct device_board *device, u32 irq);
void DisableIrqLocal(struct device_board *device, u32 irq);

void writePLX(struct device_board *device,  unsigned address, unsigned long value);
void andPLX(struct device_board *device, unsigned address, unsigned long value);
void orPLX(struct device_board *device, unsigned address, unsigned long value);
__u32 readPLX(struct device_board *device, unsigned address);

void regdump(struct device_board *device,char * label);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
irqreturn_t device_interrupt(int irq, void *dev_id, struct pt_regs *regs);
#else
void device_interrupt(int irq, void *dev_id, struct pt_regs *regs);
#endif

// apply defines not overriden above.

#include "plx_regs.h"

#endif
