/***
*** ioctl.c
***
***  General description of this file:
***     Device driver source code for General Standards 16AO12 family of 
***     16-bit analog output boards. This file is part of the Linux
***     driver source distribution for this board.
***     
***  Copyrights (c):
***     General Standards Corporation (GSC),  2002-2006
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC. 
***
***  Platform (tested on, may work with others):
***     Linux, kernel version 2.4.x, 2.6.x, Red Hat distribution, Intel hardware.
***/
//////////////////////////////////////////////////////////////////////////
// set the following flag to trace debug messages in this module.
#ifdef DEBUG
#define TRACE_LOCAL TRUE
#endif

#ifndef __KERNEL__
#define __KERNEL__
#endif

#include "internals.h"

/* ioctl file operation: this does the bulk of the work */
int device_ioctl(struct inode *inode, struct file *fp, unsigned int num, unsigned long arg)
{
    struct device_board *device = (struct device_board *)fp->private_data;
    struct register_params regs;
    unsigned long regval,ulval;

    /* main ioctl function dispatch */
    /* 'break' at the end of branches which need to wait for the */
    /* channels ready condition, 'return' from others */

    switch (num) {
    case IOCTL_NO_COMMAND:
        return (0);

    case IOCTL_INIT_BOARD:
        msg("(%d): initializing device\n", device->minor);
        mark(0xbeef0001);
        setEvent(EVENT_INIT_COMPLETE);
        EnableIrqPlxLocal(device);
        msg(" (%d): before: IntCntrlStat - %.8x\n",device->minor, readPLX(device, INT_CTRL_STATUS));
        andLocal(device,BOARD_CTRL_REG, ~(BCR_IRQ_MASK));

        device->timeout=FALSE;
        device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
        add_timer(&device->watchdog_timer);
        //msg("(%d): init about to wait\n", device->minor);
        orLocal(device, BOARD_CTRL_REG, BCR_IRQ_INIT | (BCR_INITIALIZE));
        msg(" (%d): about to wait: IntCntrlStat - %.8x\n",device->minor, readPLX(device, INT_CTRL_STATUS));
        waitEvent(EVENT_INIT_COMPLETE);
        msg(" (%d): after wait: IntCntrlStat - %.8x\n",device->minor, readPLX(device, INT_CTRL_STATUS));
        if (signal_pending(current)) {
            errmsg("Init signal Pending error\n");
            del_timer_sync(&device->watchdog_timer);
            clearEvent(EVENT_INIT_COMPLETE);
            DisableIrqPlx(device);
            return (-ERESTARTSYS); 
        }
        if (device->timeout)
        {
            //msg("%ld %ld\n",jiffies, device->watchdog_timer.expires);
            errmsg("(%d): timeout during device initialize\n", device->minor);
            clearEvent(EVENT_INIT_COMPLETE);
            device->error = DEVICE_IOCTL_TIMEOUT;
            return (-EIO);
        }
        else
            del_timer_sync(&device->watchdog_timer);

        msg("(%d): plx int ctrl %.8X board ctrl %.8X DMA ctrl status%.8X\n",device->minor, 
            readPLX(device, INT_CTRL_STATUS), readLocal(device,BOARD_CTRL_REG),readPLX(device, DMA_CMD_STATUS));
        msg("(%d): ioctl initialize done\n", device->minor);
        mark(0xbeef0002);
        return (0);

    case IOCTL_READ_REGISTER:
        copy_from_user_ret(&regs, (void *)arg, sizeof(regs), (-EFAULT));
        if ((regs.eRegister < BOARD_CTRL_REG) || (regs.eRegister > LAST_REGISTER)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        regs.ulValue = readl(device->local_addr + regs.eRegister);
        copy_to_user_ret((void *)arg, &regs, sizeof(regs), (-EFAULT));
        return (0);

    case IOCTL_WRITE_REGISTER:
        copy_from_user_ret(&regs, (void *)arg, sizeof(regs), (-EFAULT));
        if ((regs.eRegister < BOARD_CTRL_REG) || (regs.eRegister > LAST_REGISTER)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        writel(regs.ulValue, device->local_addr + regs.eRegister);
        return (0);

    case IOCTL_GET_DEVICE_ERROR:
        put_user_ret(device->error, (unsigned long *)arg, (-EFAULT));
        return (0);
        break;

    case IOCTL_SET_DMA_MODE:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if ((ulval != DMA_DISABLE && ulval != DMA_ENABLE)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        device->dmaMode=ulval;
        return (0); 

    case IOCTL_AUTOCALIBRATE:
        mark(0xbeef0003);
        //msg("(%d) -------------------------\n", device->minor);
        //msg("(%d): calibrating device\n", device->minor);
        msg("(%d): IOCTL_AUTOCALIBRATE\n", device->minor);
        EnableIrqPlxLocal(device);

        //msg("(%d): board control %.8X\n", device->minor,readlocal(device,BOARD_CTRL_REG));
        andLocal(device, BOARD_CTRL_REG, ~(BCR_IRQ_MASK));
        setEvent(EVENT_AUTOCAL_COMPLETE);

        andLocal(device, BOARD_CTRL_REG, ~BCR_CALIBRATION_MODE_MASK);

        device->timeout=FALSE;
        device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
        add_timer(&device->watchdog_timer);

        //msg("(%d): board control %.8X\n", device->minor,readlocal(device,BOARD_CTRL_REG));
        //msg("(%d): auto cal about to wait\n", device->minor);
        orLocal(device,BOARD_CTRL_REG, BCR_IRQ_AUTOCAL_COMPLETE);
        orLocal(device, BOARD_CTRL_REG, BCR_INIT_CALIBRATION);
        
        waitEvent(EVENT_AUTOCAL_COMPLETE);
        if (signal_pending(current)) {
            errmsg("(%d): signal error when calibrating device\n", device->minor);
            del_timer_sync(&device->watchdog_timer);
            DisableIrqPlx(device);
            return (-ERESTARTSYS); 
        }

        if (device->timeout) {
            errmsg("(%d): timeout when calibrating device\n", device->minor);
            clearEvent(EVENT_AUTOCAL_COMPLETE);
            device->error = DEVICE_IOCTL_TIMEOUT;
            return (-EIO);
        }
        else
            del_timer_sync(&device->watchdog_timer);

        msg("(%d): ioctl autocal done\n", device->minor);
        mark(0xbeef0004);
        return (0);

    case IOCTL_PROGRAM_RATE_GEN:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if ((ulval < 0) || (ulval > MAX_RATE_GEN)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        writeLocal(device,RATE_CTRL_REG, ulval);
        break;

    case IOCTL_CHANNEL_ENABLE:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if ((ulval < 0) || (ulval > CHAN_ALL_ENABLE)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        writeLocal(device, CHANNEL_SELECTION_REG, ulval);
        break;

    case IOCTL_SET_OUT_BUFFER_SIZE:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if (ulval <0 || ulval > OUT_BUFFER_SIZE_131072) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        andLocal(device, BUFFER_OPS_REG, ~BOR_BUFFER_SIZE_MASK);
        orLocal(device, BUFFER_OPS_REG, ulval);
        return (0);

    case IOCTL_GET_BUFFER_STATUS:
        put_user_ret(readLocal(device, BUFFER_OPS_REG)&BUFFER_STATUS_MASK, (unsigned long *)arg, (-EFAULT));
        return (0);

    case IOCTL_ENABLE_CLK:
        orLocal(device, BUFFER_OPS_REG, BOR_ENABLE_CLOCK);
        return (0);
        break;

    case IOCTL_DISABLE_CLK:
        andLocal(device, BUFFER_OPS_REG, ~(BOR_ENABLE_CLOCK));
        return (0);
        break;

    case IOCTL_SELECT_DATA_FORMAT:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if ((ulval != TWOS_COMP) && (ulval != OFFSET_BINARY)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        if (ulval==OFFSET_BINARY)
        {
            orLocal(device, BOARD_CTRL_REG, BCR_OFFSET_BINARY);
        }
        else
        {
            andLocal(device, BOARD_CTRL_REG, ~(BCR_OFFSET_BINARY));
        }

        return (0);

    case IOCTL_SELECT_SAMPLING_MODE:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if ((ulval != CONT_MODE) && (ulval != BURST_MODE)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        if (ulval==BURST_MODE)
        {
            orLocal(device, BOARD_CTRL_REG, BCR_BURST_ENABLED);
        }
        else
        {
            andLocal(device, BOARD_CTRL_REG, ~(BCR_BURST_ENABLED));
        }
        return (0);

    case IOCTL_GET_BURSTING_STATUS:
        put_user_ret(readLocal(device, BOARD_CTRL_REG)&BCR_BURST_READY, (unsigned long *)arg, (-EFAULT));
        return (0);

    case IOCTL_BURST_TRIGGER:
        orLocal(device, BOARD_CTRL_REG, BCR_BURST_TRIGGER);
        return (0);

    case IOCTL_ENABLE_REMOTE_GND_SENSE:
        orLocal(device, BOARD_CTRL_REG, BCR_REMOTE_GROUND_SENSE);
        return (0);

    case IOCTL_DISABLE_REMOTE_GND_SENSE:
        andLocal(device, BOARD_CTRL_REG, ~BCR_REMOTE_GROUND_SENSE);
        return (0);

    case IOCTL_SELECT_OUT_CLKING_MODE:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if ((ulval != SEQUENTIAL) && (ulval != SIMULTANEOUS)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }

        if (ulval==SIMULTANEOUS)
        {
            orLocal(device, BOARD_CTRL_REG, BCR_SIMULTANEOUS_OUTPUTS);
        }
        else
        {
            andLocal(device, BOARD_CTRL_REG, ~(BCR_SIMULTANEOUS_OUTPUTS));
        }
        return (0);

    case IOCTL_SELECT_CLK_SOURCE:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if ((ulval != EXTERNAL) && (ulval != INTERNAL)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        if (ulval==EXTERNAL)
        {
            orLocal(device, BUFFER_OPS_REG, BOR_EXTERNAL_CLOCK);
        }
        else
        {
            andLocal(device, BUFFER_OPS_REG, ~(BOR_EXTERNAL_CLOCK));
        }
        return (0);

    case IOCTL_GET_CLK_STATUS:
        put_user_ret(readLocal(device, BUFFER_OPS_REG)&BOR_CLOCK_READY, (unsigned long *)arg, (-EFAULT));
        return (0);         

    case IOCTL_SINGLE_OUTPUT_CLK_EVT:
        orLocal(device, BUFFER_OPS_REG, BOR_SW_CLOCK);
        return (0);         

    case IOCTL_SELECT_BUF_CONFIG:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if ((ulval != OPEN_BUF) && (ulval != CIRCULAR_BUF)) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        if (ulval==CIRCULAR_BUF)
        {
            orLocal(device, BUFFER_OPS_REG, CIRCULAR_BUF);
        }
        else
        {
            andLocal(device, BUFFER_OPS_REG, ~(CIRCULAR_BUF));
        }
        return (0);         

    case IOCTL_LOAD_ACCESS_REQ:
        regval = readLocal(device, BUFFER_OPS_REG);
        if ((regval & BOR_CIRCULAR_BUFFER) == 0) {
            device->error = DEVICE_INVALID_PARAMETER;
            return (-EIO);
        }
        
        // set up to wait for the IRQ.

        setEvent(EVENT_LOAD_READY);

        device->timeout=FALSE;
        device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
        add_timer(&device->watchdog_timer);

        EnableIrqPlxLocal(device);
        orLocal(device, BUFFER_OPS_REG, BOR_LOAD_REQUEST | BOR_LOAD_READY);
        waitEvent(EVENT_LOAD_READY);
       if (signal_pending(current)) {
           errmsg("Signal error waiting for load\n");
            del_timer_sync(&device->watchdog_timer);
            DisableIrqPlx(device);
            return (-ERESTARTSYS); 
        }
        if (device->timeout)
        {
            clearEvent(EVENT_LOAD_READY);
            device->error = DEVICE_IOCTL_TIMEOUT;
            device->timeout=FALSE;
            return (-EIO);
        }
        else
            del_timer_sync(&device->watchdog_timer);

        if (signal_pending(current)) {
            errmsg("Load signal Pending error\n");
            del_timer_sync(&device->watchdog_timer);
            clearEvent(EVENT_INIT_COMPLETE);
            DisableIrqPlx(device);
            return (-ERESTARTSYS); 
        }

        return (0);         

    case IOCTL_GET_CIR_BUF_STATUS:
        put_user_ret(readLocal(device, BUFFER_OPS_REG)&BOR_LOAD_READY, (unsigned long *)arg, (-EFAULT));
        return (0);         

    case IOCTL_CLEAR_BUFFER:
        orLocal(device, BUFFER_OPS_REG, BOR_CLEAR_BUFFER);
        return (0);

    case IOCTL_SET_TIMEOUT:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        device->timeout_seconds = ulval;
        return (0);

    case IOCTL_SET_BIG_ENDIAN:
        get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
        if (ulval)
        {
            orPLX(device, BIG_LITTLE_ENDIAN_DESC, DMA_0_BIG_ENDIAN|DMA_1_BIG_ENDIAN);
        }
        else
        {
            andPLX(device, BIG_LITTLE_ENDIAN_DESC, ~(DMA_0_BIG_ENDIAN|DMA_1_BIG_ENDIAN));
        };
        return (0);

    default:
        return (-EINVAL);
    }
    return (0);         
}
