/***
*** write.c
***
***  General description of this file:
***     Device driver source code for General Standards 16AO12 family of 
***     16-bit analog output boards. This file is part of the Linux
***     driver source distribution for this board.
***     
***  Copyrights (c):
***     General Standards Corporation (GSC),  2002-2006
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC. 
***
***  Platform (tested on, may work with others):
***     Linux, kernel version 2.4.x, 2.6.x, Red Hat distribution, Intel hardware.
***/

//////////////////////////////////////////////////////////////////////////
// set the following flag to trace debug messages in this module.
#ifdef DEBUG
#define TRACE_LOCAL TRUE
#endif

#include "internals.h"

/************************************************************************/
/* write operation: either polled or uses PLX DMA on CH0                 */
/************************************************************************/
int device_write(struct file *fp, const char *buf, size_t size, loff_t * lt)
{
    struct device_board *device = (struct device_board *)fp->private_data;
    int nsamples = size / sizeof(long);
    int bor_reg;
    unsigned long regval;

    /* verify parameters */
    if (nsamples <= 0) {
        msg("(%d): device_write - zero sample return\n", device->minor);
        return (0);
    }

    if (!access_ok(VERIFY_READ, (unsigned long)buf, nsamples * sizeof(long))) {
        errmsg("(%d): device_write - access not OK\n", device->minor);
        return (-EFAULT);
    }

    // if in circular mode, make sure the buffer is ready.  If busy for non-blocking,
    // return immediately.  For blocking wait for the buffer to come around.

    bor_reg = readLocal(device,BUFFER_OPS_REG);
    if ((fp->f_flags & O_NONBLOCK) && (bor_reg&BOR_CIRCULAR_BUFFER))
    {
        device->error = DEVICE_SUCCESS;
        return (-EAGAIN);
    }

    // if circular, wait for buffer ready

    EnableIrqPlxLocal(device);

    andLocal(device,BOARD_CTRL_REG, ~(BCR_IRQ_MASK));
    orLocal(device,BOARD_CTRL_REG, BCR_IRQ_LOAD_READY);

    if (bor_reg&BOR_CIRCULAR_BUFFER)
    {
        setEvent(EVENT_LOAD_READY);

        // start wait for buffer

        device->timeout=FALSE;
        device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
        add_timer(&device->watchdog_timer);
        msg("device_write about to wait for load ready\n");
        orLocal(device, BUFFER_OPS_REG, BOR_LOAD_REQUEST);

        waitEvent(EVENT_LOAD_READY);
        if (signal_pending(current)) {
            errmsg("signal error waiting for load ready\n");
            del_timer_sync(&device->watchdog_timer);
            DisableIrqPlx(device);
            return (-ERESTARTSYS); 
        }
        if (device->timeout)
        {
            errmsg("(%d): timeout during wait for load ready\n", device->minor);
            clearEvent(EVENT_LOAD_READY);
            device->error = DEVICE_IOCTL_TIMEOUT;
            return (-EIO);
        }
        else
            del_timer_sync(&device->watchdog_timer);

        //msg("device_write after wait for load ready\n");
    }
    else // wait for space in the buffer, as in the low-quarter state.
    {
        // disable wait for buffer open.
        andLocal(device, BOARD_CTRL_REG, ~(BCR_IRQ_MASK));

        // enable the buffer low int 
        andLocal(device, BOARD_CTRL_REG, ~(BCR_IRQ_MASK));
        orLocal(device, BOARD_CTRL_REG, BCR_IRQ_OUT_BUFFER_LOW_QUARTER);
        setEvent(EVENT_OUT_BUFFER_LOW_QUARTER);

        regval = readLocal(device,BUFFER_OPS_REG);

        if (!(regval&BOR_BUFFER_LOW_QUARTER))
        {
            device->timeout=FALSE;
            device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
            add_timer(&device->watchdog_timer);
            msg("device_write about to wait for buffer low quarter\n");
            EnableIrqPlxLocal(device);
            msg("Wait for buffer space irq\n");
            waitEvent(EVENT_OUT_BUFFER_LOW_QUARTER);
            if (signal_pending(current)) {
                errmsg("Signal error waiting for low quarter\n");
                del_timer_sync(&device->watchdog_timer);
                DisableIrqPlx(device);
                return (-ERESTARTSYS); 
            }
            if (device->timeout)
            {
                errmsg("(%d): timeout during wait for buffer low quarter\n", device->minor);
                clearEvent(EVENT_OUT_BUFFER_LOW_QUARTER);
                device->error = DEVICE_IOCTL_TIMEOUT;
                return (-EIO);
            }
            else
                del_timer_sync(&device->watchdog_timer);
        }
        else
        {
            msg("device_write already at low quarter\n");
            // disable wait for buffer low quarter.
            andLocal(device,BOARD_CTRL_REG, ~(BCR_IRQ_MASK));
            DisableIrqPlx(device);
        }
    }

    // now ready to transfer data
    if (device->dmaMode == DMA_ENABLE) { // use DMA for the write.
        //  int time_remain;
        msg("(%d): device_write_dma - %p %p %d\n", device->minor,device->dma_data[TX].intermediateBuffer,buf,size);
        setEvent(EVENT_DMA_PENDING);
        EnableIrqPlxDMA0(device);
        __copy_from_user(device->dma_data[TX].intermediateBuffer,buf,size);

        writePLX(device, DMA_CH_0_MODE, NON_DEMAND_DMA_MODE);
        writePLX(device, DMA_CH_0_PCI_ADDR, device->dma_data[TX].intermediatePhysicalAddr);
        writePLX(device, DMA_CH_0_LOCAL_ADDR, OUTPUT_BUF_REG*sizeof(__u32));
        writePLX(device, DMA_CH_0_TRANS_BYTE_CNT, size);

        writePLX(device, DMA_CH_0_DESC_PTR, 0);
        writePLX(device, DMA_MODE_ARBITRATION, 0);
        writePLX(device, DMA_THRESHOLD_REG, 0);

        andPLX(device, DMA_CMD_STATUS, STOP_DMA_CMD_0_MASK);

        orPLX(device, DMA_CMD_STATUS, CH0_DMA_ENABLE_MASK | CH0_DMA_CLEAR_IRQ_MASK);

        //msg("(%d): device_write_dma - about to sleep...\n", device->minor);
        device->timeout=FALSE;
        device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
        add_timer(&device->watchdog_timer);

        orPLX(device, DMA_CMD_STATUS, CH0_DMA_START_MASK);
        waitEvent(EVENT_DMA_PENDING);
        if (signal_pending(current)) {
            errmsg("Dma signal error.\n");
            del_timer_sync(&device->watchdog_timer);
            DisableIrqPlx(device);
            clearEvent(EVENT_DMA_PENDING);
            return (-ERESTARTSYS); 
        }
        if (device->timeout)
        {
            errmsg("(%d): device_write_dma - timeout...\n", device->minor);
            //errmsg(" (%d): In progress: %d PXL int/ctrl %.8X BC %.8X DMA %.8X\n", device->minor,device->dmainprogress, readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readl(DMACmdStatus(device)));
            clearEvent(EVENT_DMA_PENDING);

            return (-EAGAIN);
        }
        else
            del_timer_sync(&device->watchdog_timer);
#ifdef DEBUG
        //  if (time_remain< 1000)
        //      msg(" %d ", time_remain);
#endif

    }
    else // pio
    {
        int i;
        unsigned long *dataPtr;
        unsigned long data;

        dataPtr=(unsigned long *) buf;
        clearEvent(EVENT_DMA_PENDING);
        //msg("(%d): device_write PIO...\n",device->minor);
        for(i=0;i<nsamples;i++)
        {
            // this can probably be refined...
            __copy_from_user(&data,dataPtr+i,sizeof(unsigned long));
            writeLocal(device,OUTPUT_BUF_REG, data);
        }
    }
    return size;
}

